#pragma once
#include <boost/statechart/state_machine.hpp>
#include <boost/statechart/event.hpp>
#include <boost/statechart/transition.hpp>
#include <boost/statechart/state.hpp>
#include <boost/statechart/simple_state.hpp>
#include <boost/statechart/custom_reaction.hpp>
#include <boost/mpl/list.hpp>
#include "communication/bridge/connection_lifecycle/i_connector.h"
#include "communication/bridge/connection_lifecycle/i_monitor.h"
#include "communication/bridge/connection_lifecycle/i_connection_state_machine.h"
#include <kr2_program_api/internals/api/console.h>
#include <iostream>
#include <functional>
#include "spdlog/spdlog.h"

namespace spin
{
    namespace sc = boost::statechart;
    namespace mpl = boost::mpl;

    // States
    struct Idle;
    struct TryingToConnect;
    struct ConsolidatingConnection;
    struct Connected;

    // Events
    struct EvConnect : sc::event<EvConnect> { };
    struct EvDisconnect : sc::event<EvDisconnect> { };
    struct EvConnectionEstablished : sc::event<EvConnectionEstablished> { };
    struct EvWatchdogDied : sc::event<EvWatchdogDied> { };
    struct EvWatchdogKicked : sc::event<EvWatchdogKicked> { };
    struct EvCommunicationFailed : sc::event<EvCommunicationFailed> { };

    struct ConnectionStateMachine : sc::state_machine<ConnectionStateMachine, Idle>, public IConnectionStateMachine
    {

        ConnectionStateMachine(const std::shared_ptr<IConnector> connector, const std::shared_ptr<IMonitor> monitor) : connector_(connector), monitor_(monitor)
        {
        }
        virtual ~ConnectionStateMachine()
        {
            terminate();
        }

        template <class Event>
        void monitorHeartbeat(const Event &)
        {
            monitor_->monitorHeartbeat();
        }

        template <class Event>
        void stopMonitoring(const Event &)
        {
            monitor_->stopMonitoring();
        }

        bool isBridgeConnected() const override
        {
            return state_cast<const Connected *>() != nullptr;
        }

        void handleEvent(IConnectionStateMachine::Event event) override
        {
            switch (event)
            {
            case IConnectionStateMachine::Event::Connect:
                process_event(EvConnect());
                break;
            case IConnectionStateMachine::Event::Disconnect:
                process_event(EvDisconnect());
                break;
            case IConnectionStateMachine::Event::ConnectionEstablished:
                process_event(EvConnectionEstablished());
                break;
            case IConnectionStateMachine::Event::WatchdogDied:
                process_event(EvWatchdogDied());
                break;
            case IConnectionStateMachine::Event::WatchdogKicked:
                process_event(EvWatchdogKicked());
                break;
            case IConnectionStateMachine::Event::CommunicationFailed:
                process_event(EvCommunicationFailed());
                break;
            default:
                throw std::runtime_error("ConnectionStateMachine:handleEvent: "
                                         "Unexpected event");
            }
        }

        std::shared_ptr<IConnector> connector_;
        std::shared_ptr<IMonitor> monitor_;
    };

    struct Idle : sc::state<Idle, ConnectionStateMachine>
    {
        Idle(my_context ctx) : my_base(ctx)
        {
            SPDLOG_INFO("Entering state: Idle");
        }
        ~Idle()
        {
            SPDLOG_INFO("Leaving state: Idle");
        }

        using reactions = mpl::list<sc::transition<EvConnect, TryingToConnect>,
                                    sc::custom_reaction<EvConnectionEstablished>,
                                    sc::custom_reaction<EvCommunicationFailed>,
                                    sc::custom_reaction<EvWatchdogKicked>>;

        template <class Event>
        sc::result react(Event const &event)
        {
            throw std::runtime_error("ConnectionStateMachine:Idle:react: Unexpected event");
        }
    };

    struct TryingToConnect : sc::state<TryingToConnect, ConnectionStateMachine>
    {
        TryingToConnect(my_context ctx) : my_base(ctx)
        {
            SPDLOG_INFO("Entering state: TryingToConnect");
            outermost_context().connector_->startConnecting();
        }

        ~TryingToConnect()
        {
            SPDLOG_INFO("Leaving state: TryingToConnect");
            outermost_context().connector_->stopConnecting();
        }

        using reactions = mpl::list<
            sc::transition<EvConnect, TryingToConnect>,
            sc::transition<EvDisconnect, Idle>,
            sc::transition<EvConnectionEstablished, ConsolidatingConnection, ConnectionStateMachine, &ConnectionStateMachine::monitorHeartbeat>>;
    };

    struct ConsolidatingConnection : sc::simple_state<ConsolidatingConnection, ConnectionStateMachine>
    {
        ConsolidatingConnection()
        {
            SPDLOG_INFO("Entering state: ConsolidatingConnection");
            kickCounter.reset();
        }

        ~ConsolidatingConnection()
        {
            SPDLOG_INFO("Leaving state: ConsolidatingConnection");
        }

        using reactions = mpl::list<
            sc::transition<EvConnect, TryingToConnect, ConnectionStateMachine, &ConnectionStateMachine::stopMonitoring>,
            sc::transition<EvDisconnect, Idle, ConnectionStateMachine, &ConnectionStateMachine::stopMonitoring>,
            sc::transition<EvWatchdogDied, TryingToConnect, ConnectionStateMachine, &ConnectionStateMachine::stopMonitoring>,
            sc::transition<EvCommunicationFailed, TryingToConnect, ConnectionStateMachine, &ConnectionStateMachine::stopMonitoring>,
            sc::custom_reaction<EvWatchdogKicked>>;

        sc::result react(const EvWatchdogKicked &)
        {
            if (kickCounter.isDoneCounting())
            {
                return transit<Connected>();
            }
            else
            {
                kickCounter.increment();
                return discard_event();
            }
        }
        struct KickCounter
        {
            int counter;

            void increment() { ++counter; }

            bool isDoneCounting() { return counter >= 3; }

            bool isNotDoneCounting() { return !isDoneCounting(); }

            void reset() { counter = 0; }
        };
        KickCounter kickCounter;
    };

    struct Connected : sc::state<Connected, ConnectionStateMachine>
    {
        Connected(my_context ctx) : my_base(ctx)
        {
            SPDLOG_INFO("Entering state: Connected");
            outermost_context().monitor_->monitorEverything();
        }

        ~Connected()
        {
            SPDLOG_INFO("Leaving state: Connected");
            outermost_context().monitor_->stopMonitoring();
        }

        using reactions = mpl::list<
            sc::transition<EvConnect, TryingToConnect>,
            sc::transition<EvDisconnect, Idle>,
            sc::transition<EvWatchdogDied, TryingToConnect>,
            sc::transition<EvCommunicationFailed, TryingToConnect>,
            sc::custom_reaction<EvConnectionEstablished>>;

        sc::result react(EvConnectionEstablished const &event)
        {
            throw std::runtime_error("ConnectionStateMachine:Connected:react: Unexpected event");
        }
    };
} // namespace spin